/**
Wheel mouse taken from http://wiki.processing.org/index.php/Wheel_mouse
@author Rick Companje
*/

import java.awt.event.*;
//import controlP5.*;


PImage lincoln;
PImage newLincoln;
int scale;
boolean distort, hide, saved;
DraggableBall redballin, greenballin, blueballin, orangeballin, purpleballin,
    redballout, greenballout, blueballout, orangeballout, purpleballout;
PVector center;
//PVector z1,z2,z3,w1,w2,w3;
MoebiusFun moeb;
FixFour frcn;
FixThree fix3;
String imagepath;
boolean locked, pause;
float redxOffset, redyOffset, greenxOffset, greenyOffset, 
    bluexOffset, blueyOffset, orangexOffset, orangeyOffset,
    purplexOffset, purpleyOffset;
int maxsize;
float scaleup = 1.05;
float scaledown = 1/scaleup;
File lastFile;
String funID = "";

   
void setup() {
    saved = true;
    pause = true;
    scale = 200;
    maxsize = 1024;
    distort = false;
    hide = false;
    locked = false;
  
    size(maxsize,maxsize);
    newLincoln = new PImage(width, height);
    lincoln = new PImage(0,0);

    center = new PVector(width/2, height/2);
    //frcn = new FourCorners();
    moeb = new MoebiusFun(new PVector(1,0),
			  new PVector(0,0),
			  new PVector(0,0),
			  new PVector(1,0));
    if ( funID == "" ){ funID = "moeb"; }
    /*lincoln = lincoln.get(200,100,width,height);*/

    addMouseWheelListener(new MouseWheelListener() { 
	    public void mouseWheelMoved(MouseWheelEvent mwe) { 
		mouseWheel(mwe.getWheelRotation());
	    }}); 

    initFun();
    openImageFile();  
}

void openImageFile(){
    pause = true;
    selectInput("Image to load: ","setImage", lastFile);
    while ( pause ){ print(""); }
    if (lincoln.width > lincoln.height){
	lincoln.resize(maxsize,0);
    } else {
	lincoln.resize(0,maxsize);
    }
    centerLincoln();
    //println("Centered");
    rewrite();
    //println("Rewritten");
}


void centerLincoln(){
    PVector del;
    if ( lincoln.width < maxsize ){
	del = new PVector( ( maxsize - lincoln.width ) / 2, 0 );
    } else {
	del = new PVector( 0, ( maxsize - lincoln.height ) / 2 );
    }
    redballout.bpos.add( del );
    blueballout.bpos.add( del );
    greenballout.bpos.add( del );
    rewriteMoeb();
    fixDistortion();
}

void initFun(){
    background(0);
    redballout = new DraggableBall( 5, color(128, 0,0));
    redballin = new DraggableBall( 5, color(128, 0,0));
    redballin.bpos = new PVector(center.x,center.y);
    redballout.bpos = new PVector(center.x,center.y);
    orangeballout  = new DraggableBall( 5, color(192, 128, 0));
    orangeballin = new DraggableBall( 5, color(192, 128, 0));
    orangeballin.bpos = new PVector( 1.5*center.x, 0.5*center.y);
    orangeballout.bpos = new PVector( 1.5*center.x, 0.5*center.y);
    greenballout = new DraggableBall( 5, color(0,128,0));
    greenballin = new DraggableBall( 5, color(0,128,0));
    greenballin.bpos = new PVector(center.x/2,center.y/2);
    greenballout.bpos = new PVector(center.x/2,center.y/2);
    blueballout = new DraggableBall( 5, color(0,0,128));
    blueballin = new DraggableBall( 5, color(0,0,128));
    blueballin.bpos = new PVector(3*center.x/2,3*center.y/2);
    blueballout.bpos = new PVector(3*center.x/2,3*center.y/2);
    purpleballout = new DraggableBall( 5, color(128,0,128));
    purpleballin = new DraggableBall( 5, color(128,0,128));
    purpleballin.bpos = new PVector( 0.5*center.x, 1.5*center.y );
    purpleballout.bpos = new PVector( 0.5*center.x, 1.5*center.y );
}

void setImage(File fid){
  if ( fid != null ){
      //lincoln = new PImage();
      lincoln.resize(0,0);
      lincoln = loadImage(fid.getAbsolutePath()); 
      lastFile = fid;
      hide = false;
      locked = false;
  }
  pause = false;
}

void saveImage(){
  selectOutput("Select file to save to: ","imageSaver",lastFile);
}

void imageSaver(File fid){
  if ( fid != null ){
    newLincoln.save(fid.getAbsolutePath());
   lastFile = fid;
   saved = true;
  }
}


void rewrite(){
    if ( funID == "moeb" ){
	rewriteMoeb();
    } else if ( funID == "frcn" ){
	rewriteFixFour();
    } else {
	rewriteFixThree();
    }
    saved = false;
}
  
void rewriteMoeb() {
    moeb.calcCons( redballout.bpos,greenballout.bpos,blueballout.bpos,
		   redballin.bpos,greenballin.bpos,blueballin.bpos);
    //println("center = " + center);
    //println("mfun(center) = " + mfun.out(center));
    //println();
    //newLincoln = new PImage(width, height);
    for ( int x = 0; x < newLincoln.width; x++ ) {
	for ( int y = 0; y < newLincoln.height; y++ ) {
	    PVector z = new PVector(x,y);
	    PVector w = moeb.out(z);
	    color c = interpolate(lincoln, w);
	    newLincoln.set(x, y, c);
	}
    }
    /*newLincoln.resize(width,height);*/
    //image(newLincoln,0,0);
}
void rewriteFixFour() {
    frcn = new FixFour( greenballin.bpos, orangeballin.bpos,
			blueballin.bpos, purpleballin.bpos,
			redballout.bpos, redballin.bpos);
    //newLincoln = new PImage(width, height);
    for ( int x = 0; x < newLincoln.width; x++ ) {
	for ( int y = 0; y < newLincoln.height; y++ ) {
	    // PVector z = new PVector(x + redballout.bpos.x,
	    // 			    y + redballout.bpos.y);
	    PVector z = new PVector(x,y);
	    PVector w = frcn.out(z);
	    color c = interpolate(lincoln, w);
	    newLincoln.set(x, y, c);
	}
    }
    /*newLincoln.resize(width,height);*/
    //image(newLincoln,0,0);
}

void rewriteFixThree() {
    fix3 = new FixThree( greenballin.bpos, blueballin.bpos, 
			 purpleballin.bpos,
			 orangeballout.bpos, orangeballin.bpos,
			 redballout.bpos, redballin.bpos);
    // println("P(0) = " + frcn.out(center));
    // println(" frcn a0 = " + frcn.zeroPoly.a[0]);
    // println(" frcn a1 = " + frcn.zeroPoly.a[1]);
    // println(" frcn a2 = " + frcn.zeroPoly.a[2]);
    // println(" frcn a3 = " + frcn.zeroPoly.a[3]);
    // println(" frcn a4 = " + frcn.zeroPoly.a[4]);
    // println(" red = " + redballin.bpos );
    // println(" P(red) = " + frcn.zeroPoly.out( redballin.bpos ));
    // println(" green = " + greenballin.bpos );
    // println(" P(green) = " + frcn.zeroPoly.out( greenballin.bpos ));
    // println(" frcn(green) = " + frcn.out( greenballin.bpos ));
    // println();
    //newLincoln = new PImage(width, height);
    for ( int x = 0; x < newLincoln.width; x++ ) {
	for ( int y = 0; y < newLincoln.height; y++ ) {
	    PVector z = new PVector(x,y);
	    PVector w = fix3.out(z);
	    color c = interpolate(lincoln, w);
	    newLincoln.set(x, y, c);
	}
    }
    //image(newLincoln,0,0);
}




PVector toCenter( PVector z){
    PVector w0 = cAdd( z, cNeg(center));
    w0.div(scale);
    return( w0 );
}
PVector fromCenter( PVector z){
    PVector w0 = new PVector();
    w0.mult( z, scale, w0);
    return( cAdd( w0, center) );
}



void draw() {
    background(0);
    if ( distort ){
	image(newLincoln,0,0);
	if ( !hide){
	    redballout.draw();
	    greenballout.draw();
	    blueballout.draw();
	    if ( funID == "frcn" || funID == "fix3" ){
		orangeballout.draw();
		purpleballout.draw();
	    } 
	}
    } else {
	image(lincoln,0,0);
	if ( !hide ){
	    redballin.draw();
	    greenballin.draw();
	    blueballin.draw();
	    if (funID == "frcn" || funID == "fix3" ){
		orangeballin.draw();
		purpleballin.draw();
	    }
	}
    }
    if ( !hide ){
	fill(255);
	text( funID, width - 40, height - 10);
	if ( distort ){
	    if ( saved ){
		text( " distorted ", 10, height - 10);
	    } else {
		text( "*distorted*", 10, height - 10);
	    }
	}
    }
}

void mousePressed() {
    /*distort = !distort;*/
    if ( !hide ){
	if ( distort ){
	    if ( funID == "moeb" ){
		if( !redballout.mousePressed() && 
		    !greenballout.mousePressed() &&
		    !blueballout.mousePressed() ){
		    mousePressedBG();
		}
	    } else {
		if( !redballout.mousePressed() ){
		    if ( funID == "fix3" &&
			 !orangeballout.mousePressed() ){
			mousePressedBG();
		    }
		}
	    }
	} else {
	    if ( !redballin.mousePressed() &&
		 !greenballin.mousePressed() &&
		 !blueballin.mousePressed() &&
		 !orangeballin.mousePressed() &&
		 !purpleballin.mousePressed()){
		mousePressedBG();
	    }
	}
    }
}

void mouseDragged() {
  if ( !hide ){
  if ( distort ){
  redballout.mouseDragged();
  greenballout.mouseDragged();
  blueballout.mouseDragged();
  orangeballout.mouseDragged();
  purpleballout.mouseDragged();
  } else {
  redballin.mouseDragged();
  greenballin.mouseDragged();
  blueballin.mouseDragged();
  orangeballin.mouseDragged();
  purpleballin.mouseDragged();
  }
  }
  mouseDraggedBG();
}

void mouseReleased() {
    if ( !hide ){
	if ( distort ){
	    redballout.mouseReleased();
	    greenballout.mouseReleased();
	    blueballout.mouseReleased();
	    orangeballout.mouseReleased();
	    purpleballout.mouseReleased();

	    if ( funID == "frcn" ){
		greenballin.bpos = new PVector(greenballout.bpos.x,greenballout.bpos.y);
		blueballin.bpos = new PVector(blueballout.bpos.x,blueballout.bpos.y);
		orangeballin.bpos = new PVector(orangeballout.bpos.x,orangeballout.bpos.y);
		purpleballin.bpos = new PVector(purpleballout.bpos.x,purpleballout.bpos.y);
	    } else if ( funID == "fix3" ){
		greenballin.bpos = new PVector(greenballout.bpos.x,greenballout.bpos.y);
		blueballin.bpos = new PVector(blueballout.bpos.x,blueballout.bpos.y);
		purpleballin.bpos = new PVector(purpleballout.bpos.x,purpleballout.bpos.y);
	    }

	    rewrite();
	} else {
	    redballin.mouseReleased();
	    greenballin.mouseReleased();
	    blueballin.mouseReleased();
	    orangeballin.mouseReleased();
	    purpleballin.mouseReleased();
  
	    redballout.bpos = new PVector(redballin.bpos.x,redballin.bpos.y);
	    greenballout.bpos  = new PVector(greenballin.bpos.x,greenballin.bpos.y);
	    blueballout.bpos   = new PVector(blueballin.bpos.x,blueballin.bpos.y);
	    orangeballout.bpos = new PVector(orangeballin.bpos.x,orangeballin.bpos.y);
	    purpleballout.bpos = new PVector(purpleballin.bpos.x,purpleballin.bpos.y);
  
	    rewrite();
	}
    }
    mouseReleasedBG();
}

void keyPressed(){
    if ( key == ' ' ){
        distort = !distort;
    } else if ( key == 'h' ){
      hide = !hide;
    } else if ( key == 'f' ){
       fixDistortion(); 
    } else if ( key == 'q' ){
      exit();
    } else if ( key == 'o' ){
      initFun();
      openImageFile(); 
    } else if ( key == 's' ){
      saveImage(); 
    } else if ( key == 'c' ){
      toggleFun(); 
    }
}

void toggleFun(){
    if ( funID == "moeb" ){
	funID = "frcn";
    } else if ( funID == "frcn" ){
	funID = "fix3";
    } else if ( funID == "fix3" ){
	funID = "moeb";
    } 
    initFun();
    distort = false;
    rewrite();
}

void fixDistortion(){
    //println("Fixing distortion");
    lincoln.resize(width,height);
    lincoln.copy(newLincoln,0,0,width,height,0,0,width,height);
    redballin.bpos = new PVector(center.x,center.y);
    redballout.bpos = new PVector(center.x,center.y);
    greenballin.bpos = new PVector(center.x/2,center.y/2);
    greenballout.bpos = new PVector(center.x/2,center.y/2);
    blueballin.bpos = new PVector(3*center.x/2,3*center.y/2);
    blueballout.bpos = new PVector(3*center.x/2,3*center.y/2);
    orangeballin.bpos = new PVector( 1.5*center.x, 0.5*center.y);
    orangeballout.bpos = new PVector( 1.5*center.x, 0.5*center.y);
    purpleballin.bpos = new PVector( 0.5*center.x, 1.5*center.y );
    purpleballout.bpos = new PVector( 0.5*center.x, 1.5*center.y );
    rewrite();
    distort = false;
}



void mousePressedBG() {
      locked = true; 
    redxOffset = mouseX-redballout.bpos.x; 
    redyOffset = mouseY-redballout.bpos.y;
    greenxOffset = mouseX-greenballout.bpos.x; 
    greenyOffset = mouseY-greenballout.bpos.y;
    bluexOffset = mouseX-blueballout.bpos.x; 
    blueyOffset = mouseY-blueballout.bpos.y;
    orangexOffset = mouseX-orangeballout.bpos.x; 
    orangeyOffset = mouseY-orangeballout.bpos.y;
    purplexOffset = mouseX-purpleballout.bpos.x; 
    purpleyOffset = mouseY-purpleballout.bpos.y;
  }

  void mouseDraggedBG() {
    if ( locked && distort ) {
      redballout.bpos.x = mouseX-redxOffset; 
      redballout.bpos.y = mouseY-redyOffset;
      greenballout.bpos.x = mouseX-greenxOffset; 
      greenballout.bpos.y = mouseY-greenyOffset;
      blueballout.bpos.x = mouseX-bluexOffset; 
      blueballout.bpos.y = mouseY-blueyOffset;
      orangeballout.bpos.x = mouseX-orangexOffset; 
      orangeballout.bpos.y = mouseY-orangeyOffset;
      purpleballout.bpos.x = mouseX-purplexOffset; 
      purpleballout.bpos.y = mouseY-purpleyOffset;
    }}

void mouseReleasedBG() {
    locked = false;
  }
  
void mouseWheel(int delta) {
    // if ( funID == "frcn" ){
    // 	return;
    // }
    if ( !distort ){
	return;
    }
    float rat;
    if ( delta < 0 ){
	rat = scaleup;
    } else {
	rat = scaledown;
    }
    redballout.bpos.mult(rat);
    greenballout.bpos.mult(rat);
    blueballout.bpos.mult(rat);
    if ( funID == "frcn" ){
	orangeballout.bpos.mult(rat);
	purpleballout.bpos.mult(rat);
	greenballin.bpos.mult(rat);
	blueballin.bpos.mult(rat);
	orangeballin.bpos.mult(rat);
	purpleballin.bpos.mult(rat);
    } else if ( funID == "fix3" ){
	orangeballout.bpos.mult(rat);
	purpleballout.bpos.mult(rat);
	greenballin.bpos.mult(rat);
	blueballin.bpos.mult(rat);
	purpleballin.bpos.mult(rat);
    }
    rewrite();
}
 
